import re
from abc import ABC, abstractmethod


class StreamMatcher(ABC):
    """
    Abstract base class for stream matchers.
    """

    @abstractmethod
    def match(self, item):
        """
        Return ``True`` if the matcher matches an item, otherwise ``False``.
        """

    @abstractmethod
    def args_str(self):
        """
        Return an args string for the repr.
        """

    def __call__(self, item):
        return self.match(item)

    def __str__(self):
        return '{}({})'.format(type(self).__name__, self.args_str())

    def __repr__(self):
        return str(self)


def to_matcher(matcher_factory, obj):
    """
    Turn an object into a :class:`StreamMatcher` unless it already is one.

    :param matcher_factory: A callable capable of turning `obj` into a
        :class:`StreamMatcher`.
    :param obj: A :class:`StreamMatcher` or an object to turn into one.

    :returns: :class:`StreamMatcher`
    """
    return obj if isinstance(obj, StreamMatcher) else matcher_factory(obj)


class CombinationMatcher(StreamMatcher):
    """
    Matcher that combines multiple input matchers.
    """
    def __init__(self, *matchers):
        self._matchers = matchers

    @classmethod
    def by_equality(cls, *expected_items):
        """
        Construct an instance of this combination matcher from a list of
        expected items and/or StreamMatcher instances.
        """
        return cls(*(to_matcher(EqualsMatcher, i) for i in expected_items))

    @classmethod
    def by_regex(cls, *patterns):
        """
        Construct an instance of this combination matcher from a list of
        regex patterns and/or StreamMatcher instances.
        """
        return cls(*(to_matcher(RegexMatcher, p) for p in patterns))


class OrderedMatcher(CombinationMatcher):
    """
    Matcher that takes a list of matchers, and uses one after the next after
    each has a successful match. Returns True ("matches") on the final match.

    **Note:** This is a *stateful* matcher. Once it has done its matching,
    you'll need to create a new instance.
    """
    def __init__(self, *matchers):
        super().__init__(*matchers)
        self._position = 0

    def match(self, item):
        """
        Return ``True`` if the expected matchers are matched in the expected
        order, otherwise ``False``.
        """
        if self._position == len(self._matchers):
            raise RuntimeError('Matcher exhausted, no more matchers to use')

        matcher = self._matchers[self._position]
        if matcher(item):
            self._position += 1

        if self._position == len(self._matchers):
            # All patterns have been matched
            return True

        return False

    def args_str(self):
        """
        Return an args string for the repr.
        """
        matched = [str(m) for m in self._matchers[:self._position]]
        unmatched = [str(m) for m in self._matchers[self._position:]]
        return 'matched=[{}], unmatched=[{}]'.format(
            ', '.join(matched), ', '.join(unmatched))


class UnorderedMatcher(CombinationMatcher):
    """
    Matcher that takes a list of matchers, and matches each one to an item.
    Each item is tested against each unmatched matcher until a match is found
    or all unmatched matchers are checked. Returns True ("matches") on the
    final match.

    .. note::

        This is a *stateful* matcher. Once it has done its matching,
        you'll need to create a new instance.
    """
    def __init__(self, *matchers):
        super().__init__(*matchers)
        self._used_matchers = []

    @property
    def _unused_matchers(self):
        return [m for m in self._matchers if m not in self._used_matchers]

    def match(self, item):
        """
        Return ``True`` if the expected matchers are matched in any order,
        otherwise ``False``.
        """
        if not self._unused_matchers:
            raise RuntimeError('Matcher exhausted, no more matchers to use')

        for matcher in self._unused_matchers:
            if matcher(item):
                self._used_matchers.append(matcher)
                break

        if not self._unused_matchers:
            # All patterns have been matched
            return True

        return False

    def args_str(self):
        """
        Return an args string for the repr.
        """
        matched = [str(m) for m in self._used_matchers]
        unmatched = [str(m) for m in self._unused_matchers]
        return 'matched=[{}], unmatched=[{}]'.format(
            ', '.join(matched), ', '.join(unmatched))


class EqualsMatcher(StreamMatcher):
    """
    Matcher that matches items by equality.
    """
    def __init__(self, expected_item):
        self._expected_item = expected_item

    def match(self, item):
        """
        Return ``True`` if the item matches the expected value exactly,
        otherwise ``False``.
        """
        return item == self._expected_item

    def args_str(self):
        """
        Return an args string for the repr.
        """
        return repr(self._expected_item)


class RegexMatcher(StreamMatcher):
    """
    Matcher that matches items by regex pattern.
    """
    def __init__(self, pattern):
        self._regex = re.compile(pattern)

    def match(self, item):
        """
        Return ``True`` if the item matches the expected regex, otherwise
        ``False``.
        """
        return self._regex.search(item) is not None

    def args_str(self):
        """
        Return an args string for the repr.
        """
        return repr(self._regex.pattern)



"""
RabbitMQ container definition.
"""

from seaworthy.definitions import ContainerDefinition
from seaworthy.utils import output_lines


def _parse_rabbitmq_user(user_line):
    user, tags = user_line.split('\t', 1)
    tags = tags.strip('[]').split(', ')
    return (user, tags)


class RabbitMQContainer(ContainerDefinition):
    """
    RabbitMQ container definition.

    .. todo::
       Write more docs.
    """

    # For some reason this container is slower to start through seaworthy than
    # with a plain `docker run`, so give it a bit more time to get going. :-(
    WAIT_TIMEOUT = 20.0

    DEFAULT_NAME = 'rabbitmq'
    DEFAULT_IMAGE = 'rabbitmq:alpine'
    DEFAULT_WAIT_PATTERNS = (r'Server startup complete',)

    DEFAULT_VHOST = '/vhost'
    DEFAULT_USER = 'user'
    DEFAULT_PASSWORD = 'password'

    def __init__(self,
                 name=DEFAULT_NAME,
                 image=DEFAULT_IMAGE,
                 wait_patterns=DEFAULT_WAIT_PATTERNS,
                 vhost=DEFAULT_VHOST,
                 user=DEFAULT_USER,
                 password=DEFAULT_PASSWORD,
                 **kwargs):
        """
        :param vhost: the name of a vhost to create at startup
        :param user: the name of a user to create at startup
        :param password: the password for the user
        """
        super().__init__(name, image, wait_patterns, **kwargs)

        self.vhost = vhost
        self.user = user
        self.password = password

    def base_kwargs(self):
        """
        Add a ``tmpfs`` entry for ``/var/lib/rabbitmq`` to avoid unnecessary
        disk I/O and ``environment`` entries for the configured vhost and user
        creds.
        """
        return {
            'environment': {
                'RABBITMQ_DEFAULT_VHOST': self.vhost,
                'RABBITMQ_DEFAULT_USER': self.user,
                'RABBITMQ_DEFAULT_PASS': self.password,
            },
            'tmpfs': {'/var/lib/rabbitmq': 'uid=100,gid=101'},
        }

    def clean(self):
        """
        Remove all data by using ``rabbitmqctl`` to eval
        ``rabbit_mnesia:reset()``.
        """
        reset_erl = 'rabbit:stop(), rabbit_mnesia:reset(), rabbit:start().'
        self.exec_rabbitmqctl('eval', [reset_erl])

    def exec_rabbitmqctl(self, command, args=[], rabbitmqctl_opts=['-q']):
        """
        Execute a ``rabbitmqctl`` command inside a running container.

        :param command: the command to run
        :param args: a list of args for the command
        :param rabbitmqctl_opts:
            a list of extra options to pass to ``rabbitmqctl``
        :returns: a tuple of the command exit code and output
        """
        cmd = ['rabbitmqctl'] + rabbitmqctl_opts + [command] + args
        return self.inner().exec_run(cmd)

    def list_vhosts(self):
        """
        Run the ``list_vhosts`` command and return a list of vhost names.
        """
        return output_lines(self.exec_rabbitmqctl('list_vhosts'))

    def list_queues(self):
        """
        Run the ``list_queues`` command (for the default vhost) and return a
        list of tuples describing the queues.

        :return:
            A list of 2-element tuples. The first element is the queue name,
            the second is the current queue size.
        """
        lines = output_lines(
            self.exec_rabbitmqctl('list_queues', ['-p', self.vhost]))
        return [tuple(line.split(None, 1)) for line in lines]

    def list_users(self):
        """
        Run the ``list_users`` command and return a list of tuples describing
        the users.

        :return:
            A list of 2-element tuples. The first element is the username, the
            second a list of tags for the user.
        """
        lines = output_lines(self.exec_rabbitmqctl('list_users'))
        return [_parse_rabbitmq_user(line) for line in lines]

    def broker_url(self):
        """ Returns a "broker URL" for use with Celery. """
        return 'amqp://{}:{}@{}/{}'.format(
            self.user, self.password, self.name, self.vhost)



from docker.models.containers import ExecResult


def output_lines(output, encoding='utf-8'):
    """
    Convert bytestring container output or the result of a container exec
    command into a sequence of unicode lines.

    :param output:
        Container output bytes or an
        :class:`docker.models.containers.ExecResult` instance.
    :param encoding: The encoding to use when converting bytes to unicode
        (default ``utf-8``).

    :returns: list[str]
    """
    if isinstance(output, ExecResult):
        _, output = output

    return output.decode(encoding).splitlines()



"""
Classes that track resource creation and removal to ensure that all resources
are namespaced and cleaned up after use.
"""

import logging

import docker
from docker import models

# This is a hack to control our generated documentation. The value of the
# attribute is ignored, only its presence or absence can be detected by the
# apigen machinery.
__apigen_inherited_members__ = None


log = logging.getLogger(__name__)


def fetch_images(client, images):
    """
    Fetch images if they aren't already present.
    """
    return [fetch_image(client, image) for image in images]


def fetch_image(client, name):
    """
    Fetch an image if it isn't already present.

    This works like ``docker pull`` and will pull the tag ``latest`` if no tag
    is specified in the image name.
    """
    try:
        image = client.images.get(name)
    except docker.errors.ImageNotFound:
        name, tag = _parse_image_tag(name)
        tag = 'latest' if tag is None else tag

        log.info("Pulling tag '{}' for image '{}'...".format(tag, name))
        image = client.images.pull(name, tag=tag)

    log.debug("Found image '{}' for tag '{}'".format(image.id, name))
    return image


def _parse_image_tag(name_tag):
    # First get the last part of the name after a '/': this removes the
    # registry which could have a ':' in it
    last_name_part = name_tag.rsplit('/', 1)[-1]

    # Then get the last part after the ':'
    last_parts = last_name_part.rsplit(':', 1)

    if len(last_parts) == 2:
        _, tag = last_parts
        return name_tag[:-(len(tag) + 1)], tag
    else:
        return name_tag, None


def _parse_volume_short_form(short_form):
    parts = short_form.split(':', 1)
    bind = parts[0]
    mode = parts[1] if len(parts) == 2 else 'rw'
    return {'bind': bind, 'mode': mode}


class _HelperBase:
    __collection_type__ = None

    def __init__(self, client, namespace):
        self.collection = self.__collection_type__(client=client)
        self.namespace = namespace

        self._model_name = self.collection.model.__name__.lower()
        self._ids = set()

    def _resource_name(self, name):
        return '{}_{}'.format(self.namespace, name)

    def _get_id_and_model(self, id_or_model):
        """
        Get both the model and ID of an object that could be an ID or a model.

        :param id_or_model:
            The object that could be an ID string or a model object.
        :param model_collection:
            The collection to which the model belongs.
        """
        if isinstance(id_or_model, self.collection.model):
            model = id_or_model
        elif isinstance(id_or_model, str):
            # Assume we have an ID string
            model = self.collection.get(id_or_model)
        else:
            raise TypeError('Unexpected type {}, expected {} or {}'.format(
                type(id_or_model), str, self.collection.model))

        return model.id, model

    def create(self, name, *args, **kwargs):
        """
        Create an instance of this resource type.
        """
        resource_name = self._resource_name(name)
        log.info(
            "Creating {} '{}'...".format(self._model_name, resource_name))
        resource = self.collection.create(*args, name=resource_name, **kwargs)
        self._ids.add(resource.id)
        return resource

    def remove(self, resource, **kwargs):
        """
        Remove an instance of this resource type.
        """
        log.info(
            "Removing {} '{}'...".format(self._model_name, resource.name))
        resource.remove(**kwargs)
        self._ids.remove(resource.id)

    def _teardown(self):
        for resource_id in self._ids.copy():
            # Check if the resource exists before trying to remove it
            try:
                resource = self.collection.get(resource_id)
            except docker.errors.NotFound:
                continue

            log.warning("{} '{}' still existed during teardown".format(
                self._model_name.title(), resource.name))

            self._teardown_remove(resource)

    def _teardown_remove(self, resource):
        # Override in subclass for different removal behaviour on teardown
        self.remove(resource)


class ContainerHelper(_HelperBase):
    """
    .. todo::

        Document this properly.
    """
    __collection_type__ = models.containers.ContainerCollection

    def __init__(self, client, namespace, image_helper, network_helper,
                 volume_helper):
        super().__init__(client, namespace)
        self._image_helper = image_helper
        self._network_helper = network_helper
        self._volume_helper = volume_helper

    def create(self, name, image, fetch_image=False, network=None, volumes={},
               **kwargs):
        """
        Create a new container.

        :param name:
            The name for the container. This will be prefixed with the
            namespace.
        :param image:
            The image tag or image object to create the container from.
        :param network:
            The network to connect the container to. The container will be
            given an alias with the ``name`` parameter. Note that, unlike the
            Docker Python client, this parameter can be a ``Network`` model
            object, and not just a network ID or name.
        :param volumes:
            A mapping of volumes to bind parameters. The keys to this mapping
            can be any of three types of objects:

            - A ``Volume`` model object
            - The name of a volume (str)
            - A path on the host to bind mount into the container (str)

            The bind parameters, i.e. the values in the mapping, can be of
            two types:

            - A full bind specifier (dict), for example
              ``{'bind': '/mnt', 'mode': 'rw'}``
            - A "short-form" bind specifier (str), for example ``/mnt:rw``
        :param fetch_image:
            Whether to attempt to pull the image if it is not found locally.
        :param kwargs:
            Other parameters to create the container with.
        """
        create_kwargs = {
            'detach': True,
        }

        # Convert network & volume models to IDs
        network = self._network_for_container(network, kwargs)
        if network is not None:
            network_id, network = (
                self._network_helper._get_id_and_model(network))
            create_kwargs['network'] = network_id

        if volumes:
            create_kwargs['volumes'] = self._volumes_for_container(volumes)

        create_kwargs.update(kwargs)

        if fetch_image:
            self._image_helper.fetch(image)

        container = super().create(name, image, **create_kwargs)

        if network is not None:
            self._connect_container_network(container, network, aliases=[name])

        return container

    def _network_for_container(self, network, create_kwargs):
        # If a network is specified use that
        if network is not None:
            return network

        # If 'network_mode' is used or networking is disabled, don't handle
        # networking.
        if (create_kwargs.get('network_mode') is not None or
                create_kwargs.get('network_disabled', False)):
            return None

        # Else, use the default network
        return self._network_helper.get_default()

    def _volumes_for_container(self, volumes):
        create_volumes = {}
        for vol, opts in volumes.items():
            try:
                vol_id, _ = self._volume_helper._get_id_and_model(vol)
            except docker.errors.NotFound:
                # Assume this is a bind if we can't find the ID
                vol_id = vol

            if vol_id in create_volumes:
                raise ValueError(
                    "Volume '{}' specified more than once".format(vol_id))

            # Short form of opts
            if isinstance(opts, str):
                opts = _parse_volume_short_form(opts)
            # Else assume long form

            create_volumes[vol_id] = opts
        return create_volumes

    def _connect_container_network(self, container, network, **connect_kwargs):
        # FIXME: Hack to make sure the container has the right network aliases.
        # Only the low-level Docker client API allows us to specify endpoint
        # aliases at container creation time:
        # https://docker-py.readthedocs.io/en/stable/api.html#docker.api.container.ContainerApiMixin.create_container
        # If we don't specify a network when the container is created then the
        # default bridge network is attached which we don't want, so we
        # reattach our custom network as that allows specifying aliases.
        network.disconnect(container)
        network.connect(container, **connect_kwargs)
        # Reload the container data to get the new network setup
        container.reload()
        # We could also reload the network data to update the containers that
        # are connected to it but that listing doesn't include containers that
        # have been created and connected but not yet started. :-/

    def remove(self, container, force=True, volumes=True):
        """
        Remove a container.

        :param container: The container to remove.
        :param force:
            Whether to force the removal of the container, even if it is
            running. Note that this defaults to True, unlike the Docker
            default.
        :param volumes:
            Whether to remove any volumes that were created implicitly with
            this container, i.e. any volumes that were created due to
            ``VOLUME`` directives in the Dockerfile. External volumes that were
            manually created will not be removed. Note that this defaults to
            True, unlike the Docker default (where the equivalent parameter,
            ``v``, defaults to False).
        """
        super().remove(container, force=force, v=volumes)

    def _teardown_remove(self, container):
        self.remove(container, force=True)


class ImageHelper:
    """
    .. todo::

        Document this properly.
    """
    def __init__(self, client):
        self.collection = client.images

    def fetch(self, tag):
        """
        Fetch this image if it isn't already present.
        """
        return fetch_image(self.collection.client, tag)


class NetworkHelper(_HelperBase):
    """
    .. todo::

        Document this properly.
    """
    __collection_type__ = models.networks.NetworkCollection

    def __init__(self, client, namespace):
        super().__init__(client, namespace)
        self._default_network = None

    def _teardown(self):
        # Remove the default network
        if self._default_network is not None:
            self.remove(self._default_network)
            self._default_network = None

        # Remove all other networks
        super()._teardown()

    def get_default(self, create=True):
        """
        Get the default bridge network that containers are connected to if no
        other network options are specified.

        :param create:
            Whether or not to create the network if it doesn't already exist.
        """
        if self._default_network is None and create:
            log.debug("Creating default network...")
            self._default_network = self.create('default', driver='bridge')

        return self._default_network

    def create(self, name, check_duplicate=True, **kwargs):
        """
        Create a new network.

        :param name:
            The name for the network. This will be prefixed with the namespace.
        :param check_duplicate:
            Whether or not to check for networks with the same name. Docker
            allows the creation of multiple networks with the same name (unlike
            containers). This seems to cause problems sometimes for some reason
            (?). The Docker Python client _claims_ (as of 2.5.1) that
            ``check_duplicate`` defaults to True but it actually doesn't. We
            default it to True ourselves here.
        :param kwargs:
            Other parameters to create the network with.
        """
        return super().create(name, check_duplicate=check_duplicate, **kwargs)


class VolumeHelper(_HelperBase):
    """
    .. todo::

        Document this properly.
    """
    __collection_type__ = models.volumes.VolumeCollection

    def create(self, name, **kwargs):
        """
        Create a new volume.

        :param name:
            The name for the volume. This will be prefixed with the namespace.
        :param kwargs:
            Other parameters to create the volume with.
        """
        return super().create(name, **kwargs)


class DockerHelper:
    """
    .. todo::

        Document this properly.
    """

    def __init__(self, namespace='test', client=None):
        self._namespace = namespace
        if client is None:
            client = docker.client.from_env()
        self._client = client

        self.images = ImageHelper(self._client)
        self.networks = NetworkHelper(self._client, namespace)
        self.volumes = VolumeHelper(self._client, namespace)
        self.containers = ContainerHelper(
            self._client, namespace, self.images, self.networks, self.volumes)

    def _helper_for_model(self, model_type):
        """
        Get the helper for a given type of Docker model. For use by resource
        definitions.
        """
        if model_type is models.containers.Container:
            return self.containers
        if model_type is models.images.Image:
            return self.images
        if model_type is models.networks.Network:
            return self.networks
        if model_type is models.volumes.Volume:
            return self.volumes

        raise ValueError('Unknown model type {}'.format(model_type))

    def teardown(self):
        """
        Clean up all resources when we're done with them.
        """
        self.containers._teardown()
        self.networks._teardown()
        self.volumes._teardown()

        # We need to close the underlying APIClient explicitly to avoid
        # ResourceWarnings from unclosed HTTP connections.
        self._client.api.close()



import unittest

from testtools.assertions import assert_that
from testtools.matchers import Not

from seaworthy.ps import PsRow, PsTree
from seaworthy.testtools import MatchesPsTree


class TestMatchesPsTree(unittest.TestCase):
    def test_minimal_tree_matches(self):
        """
        MatchesPsTree can match a single-process tree.
        """
        ps_tree = PsTree(PsRow(1, 0, 'root', 'tini -- echo "hi"'))

        matcher = MatchesPsTree('root', 'tini -- echo "hi"')
        assert matcher.match(ps_tree) is None

        matcher = MatchesPsTree('root', 'tini -- echo "hi"', 1)
        assert matcher.match(ps_tree) is None

        matcher = MatchesPsTree('root', 'tini -- echo "hi"', ppid=0)
        assert matcher.match(ps_tree) is None

        matcher = MatchesPsTree('root', 'tini -- echo "hi"', 1, 0)
        assert matcher.match(ps_tree) is None

        matcher = MatchesPsTree('root', 'tini -- echo "hi"', 1, 0, children=[])
        assert matcher.match(ps_tree) is None

    def test_minimal_tree_mismatches(self):
        """
        MatchesPsTree can detect a non-matching single-process tree.
        """
        ps_tree = PsTree(PsRow(1, 0, 'root', 'tini -- true'))

        matcher = MatchesPsTree('tuber', 'tini -- true')
        mismatch = matcher.match(ps_tree)
        assert "'root' != 'tuber': ruser" in mismatch.describe()

        matcher = MatchesPsTree('root', 'tini -- false')
        mismatch = matcher.match(ps_tree)
        assert "'tini -- true' != 'tini -- false': args" in mismatch.describe()

        matcher = MatchesPsTree('tuber', 'tini -- true', pid=7)
        mismatch = matcher.match(ps_tree)
        assert "1 != 7: pid" in mismatch.describe()

        matcher = MatchesPsTree('tuber', 'tini -- true', ppid=7)
        mismatch = matcher.match(ps_tree)
        assert "0 != 7: ppid" in mismatch.describe()

    def test_nested_tree_matches(self):
        """
        MatchesPsTree can match a multi-process tree.
        """
        ps_tree = PsTree(PsRow(1, 0, 'root', 'tini -- app'), [
            PsTree(PsRow(2, 1, 'root', 'app --arg'), [
                PsTree(PsRow(3, 2, 'appuser', 'app --child1')),
                PsTree(PsRow(4, 2, 'appuser', 'app --child2')),
            ]),
            PsTree(PsRow(5, 1, 'root', 'app2 --arg'), [
                PsTree(PsRow(6, 5, 'root', 'app2 --child')),
            ]),
        ])

        # Check children in the same order.
        matcher = MatchesPsTree('root', 'tini -- app', pid=1, children=[
            MatchesPsTree('root', 'app --arg', children=[
                MatchesPsTree('appuser', 'app --child1'),
                MatchesPsTree('appuser', 'app --child2'),
            ]),
            MatchesPsTree('root', 'app2 --arg', children=[
                MatchesPsTree('root', 'app2 --child'),
            ]),
        ])
        assert matcher.match(ps_tree) is None

        # Check children in a different order.
        matcher = MatchesPsTree('root', 'tini -- app', pid=1, children=[
            MatchesPsTree('root', 'app --arg', children=[
                MatchesPsTree('appuser', 'app --child2'),
                MatchesPsTree('appuser', 'app --child1'),
            ]),
            MatchesPsTree('root', 'app2 --arg', children=[
                MatchesPsTree('root', 'app2 --child'),
            ]),
        ])
        assert matcher.match(ps_tree) is None

        # Check different children in a different order.
        matcher = MatchesPsTree('root', 'tini -- app', pid=1, children=[
            MatchesPsTree('root', 'app2 --arg', children=[
                MatchesPsTree('root', 'app2 --child'),
            ]),
            MatchesPsTree('root', 'app --arg', children=[
                MatchesPsTree('appuser', 'app --child1'),
                MatchesPsTree('appuser', 'app --child2'),
            ]),
        ])
        assert matcher.match(ps_tree) is None

    def test_nested_tree_missing_child(self):
        """
        MatchesPsTree can detect a missing child process.
        """
        ps_tree = PsTree(PsRow(1, 0, 'root', 'tini -- app'), [
            PsTree(PsRow(2, 1, 'root', 'app --arg'), [
                PsTree(PsRow(3, 2, 'appuser', 'app --child1')),
            ]),
            PsTree(PsRow(5, 1, 'root', 'app2 --arg'), [
                PsTree(PsRow(6, 5, 'root', 'app2 --child')),
            ]),
        ])

        matcher = MatchesPsTree('root', 'tini -- app', pid=1, children=[
            MatchesPsTree('root', 'app --arg', children=[
                MatchesPsTree('appuser', 'app --child1'),
                MatchesPsTree('appuser', 'app --child2'),
            ]),
            MatchesPsTree('root', 'app2 --arg', children=[
                MatchesPsTree('root', 'app2 --child'),
            ]),
        ])
        mm = matcher.match(ps_tree).describe()
        assert "mismatches in children:" in mm
        assert "There was 1 matcher left over:" in mm

    def test_nested_tree_extra_child(self):
        """
        MatchesPsTree can detect an extra child process.
        """
        ps_tree = PsTree(PsRow(1, 0, 'root', 'tini -- app'), [
            PsTree(PsRow(2, 1, 'root', 'app --arg'), [
                PsTree(PsRow(3, 2, 'appuser', 'app --child1')),
                PsTree(PsRow(4, 2, 'appuser', 'app --child3')),
                PsTree(PsRow(7, 2, 'appuser', 'app --child2')),
            ]),
            PsTree(PsRow(5, 1, 'root', 'app2 --arg'), [
                PsTree(PsRow(6, 5, 'root', 'app2 --child')),
            ]),
        ])

        matcher = MatchesPsTree('root', 'tini -- app', pid=1, children=[
            MatchesPsTree('root', 'app --arg', children=[
                MatchesPsTree('appuser', 'app --child1'),
                MatchesPsTree('appuser', 'app --child2'),
            ]),
            MatchesPsTree('root', 'app2 --arg', children=[
                MatchesPsTree('root', 'app2 --child'),
            ]),
        ])
        mm = matcher.match(ps_tree).describe()
        assert "mismatches in children:" in mm
        assert "There was 1 value left over:" in mm

    def test_nested_tree_different_child(self):
        """
        MatchesPsTree can detect a child process that is different.
        """
        ps_tree = PsTree(PsRow(1, 0, 'root', 'tini -- app'), [
            PsTree(PsRow(2, 1, 'root', 'app --arg'), [
                PsTree(PsRow(3, 2, 'appuser', 'app --child1')),
                PsTree(PsRow(4, 2, 'appuser', 'app --child3')),
            ]),
            PsTree(PsRow(5, 1, 'root', 'app2 --arg'), [
                PsTree(PsRow(6, 5, 'root', 'app2 --child')),
            ]),
        ])

        matcher = MatchesPsTree('root', 'tini -- app', pid=1, children=[
            MatchesPsTree('root', 'app --arg', children=[
                MatchesPsTree('appuser', 'app --child1'),
                MatchesPsTree('appuser', 'app --child2'),
            ]),
            MatchesPsTree('root', 'app2 --arg', children=[
                MatchesPsTree('root', 'app2 --child'),
            ]),
        ])
        mm = matcher.match(ps_tree).describe()
        assert "mismatches in children:" in mm
        assert "'app --child3' != 'app --child2': args" in mm

    def test_using_assert_that(self):
        """
        MatchesPsTree can be used with assert_that() from testtools.
        """
        pst = PsTree(PsRow(1, 0, 'root', 'tini -- app'), [
            PsTree(PsRow(2, 1, 'root', 'app --arg'), [
                PsTree(PsRow(3, 2, 'appuser', 'app --child1')),
                PsTree(PsRow(4, 2, 'appuser', 'app --child2')),
            ]),
            PsTree(PsRow(5, 1, 'root', 'app2 --arg'), [
                PsTree(PsRow(6, 5, 'root', 'app2 --child')),
            ]),
        ])

        # This passes if the MatchesPsTree matcher matches.
        assert_that(pst, MatchesPsTree('root', 'tini -- app', children=[
            MatchesPsTree('root', 'app --arg', children=[
                MatchesPsTree('appuser', 'app --child1'),
                MatchesPsTree('appuser', 'app --child2'),
            ]),
            MatchesPsTree('root', 'app2 --arg', children=[
                MatchesPsTree('root', 'app2 --child'),
            ]),
        ]))

        # This passes if the MatchesPsTree matcher does not match.
        assert_that(pst, Not(MatchesPsTree('root', 'tini -- app', children=[
            MatchesPsTree('root', 'app --arg', children=[
                MatchesPsTree('appuser', 'app --child1'),
            ]),
            MatchesPsTree('root', 'app2 --arg', children=[
                MatchesPsTree('root', 'app2 --child'),
                MatchesPsTree('appuser', 'app --child2'),
            ]),
        ])))



"""
seaworthy
~~~~~~~~~

.. todo::

   Write some API reference docs for :mod:`seaworthy`.
"""

from .helpers import DockerHelper
from .stream.logs import wait_for_logs_matching
from .utils import output_lines

__all__ = ['DockerHelper', 'output_lines', 'wait_for_logs_matching']

__version__ = '0.4.1'



"""
Redis container definition.
"""

from seaworthy.definitions import ContainerDefinition
from seaworthy.utils import output_lines


class RedisContainer(ContainerDefinition):
    """
    Redis container definition.

    .. todo::
       Write more docs.
    """

    DEFAULT_NAME = 'redis'
    DEFAULT_IMAGE = 'redis:alpine'
    DEFAULT_WAIT_PATTERNS = (r'\* Ready to accept connections',)

    def __init__(self,
                 name=DEFAULT_NAME,
                 image=DEFAULT_IMAGE,
                 wait_patterns=DEFAULT_WAIT_PATTERNS,
                 **kwargs):
        super().__init__(name, image, wait_patterns, **kwargs)

    def base_kwargs(self):
        """
        Add a ``tmpfs`` entry for ``/data`` to avoid unnecessary disk I/O.
        """
        return {'tmpfs': {'/data': 'uid=100,gid=101'}}

    def clean(self):
        """
        Remove all data by sending the ``FLUSHALL`` command.
        """
        self.exec_redis_cli('FLUSHALL')

    def exec_redis_cli(self, command, args=[], db=0, redis_cli_opts=[]):
        """
        Execute a ``redis-cli`` command inside a running container.

        :param command: the command to run
        :param args: a list of args for the command
        :param db: the db number to query (default ``0``)
        :param redis_cli_opts: a list of extra options to pass to ``redis-cli``
        :returns: a tuple of the command exit code and output
        """
        cli_opts = ['-n', str(db)] + redis_cli_opts
        cmd = ['redis-cli'] + cli_opts + [command] + [str(a) for a in args]
        return self.inner().exec_run(cmd)

    def list_keys(self, pattern='*', db=0):
        """
        Run the ``KEYS`` command and return the list of matching keys.

        :param pattern: the pattern to filter keys by (default ``*``)
        :param db: the db number to query (default ``0``)
        """
        lines = output_lines(self.exec_redis_cli('KEYS', [pattern], db=db))
        return [] if lines == [''] else lines



import threading


def stream_timeout(stream, timeout, timeout_msg=None):
    """
    Iterate over items in a streaming response from the Docker client within
    a timeout.

    :param ~docker.types.daemon.CancellableStream stream:
        Stream from the Docker client to consume items from.
    :param timeout:
        Timeout value in seconds.
    :param timeout_msg:
        Message to raise in the exception when a timeout occurs.
    """
    timed_out = threading.Event()

    def timeout_func():
        timed_out.set()
        stream.close()

    timer = threading.Timer(timeout, timeout_func)
    try:
        timer.start()
        for item in stream:
            yield item

        # A timeout looks the same as the loop ending. So we need to check a
        # flag to determine whether a timeout occurred or not.
        if timed_out.is_set():
            raise TimeoutError(timeout_msg)
    finally:
        timer.cancel()
        # Close the stream's underlying response object (if it has one) to
        # avoid potential socket leaks.
        # This method seems to have more success at preventing ResourceWarnings
        # than just stream.close() (should this be improved upstream?)
        # FIXME: Potential race condition if Timer thread closes the stream at
        # the same time we do here, but hopefully not with serious side effects
        if hasattr(stream, '_response'):
            stream._response.close()



from seaworthy.client import wait_for_response
from seaworthy.definitions import ContainerDefinition


class NginxContainer(ContainerDefinition):
    """
    Nginx container definition.

    .. todo:
        Write more docs.
    """

    DEFAULT_NAME = 'nginx'
    DEFAULT_IMAGE = 'nginx:alpine'

    def __init__(self, name=DEFAULT_NAME, image=DEFAULT_IMAGE, **kwargs):
        super().__init__(name, image, **kwargs)

    def base_kwargs(self):
        """
        Publish all exposed ports to the host.
        """
        return {'publish_all_ports': True}

    def wait_for_start(self):
        """
        Wait for Nginx to return any valid HTTP response.
        """
        wait_for_response(self.http_client(), self.wait_timeout)

    def exec_nginx(self, args):
        """
        Execute a ``nginx`` command inside a running container.

        :params args: a list of args for the command
        """
        return self.inner().exec_run(['nginx'] + args)

    def exec_signal(self, signal='reload'):
        """
        Send a signal to the Nginx master process (``nginx -s``).

        :param signal: one of: stop, quit, reopen, or reload
        """
        return self.exec_nginx(['-s', signal])



from seaworthy.stream._timeout import stream_timeout


def _last_few_log_lines(container):
    return container.logs(tail=100).decode('utf-8')


def stream_logs(container, timeout=10.0, **logs_kwargs):
    """
    Stream logs from a Docker container within a timeout.

    :param ~docker.models.containers.Container container:
        Container who's log lines to stream.
    :param timeout:
        Timeout value in seconds.
    :param logs_kwargs:
        Additional keyword arguments to pass to ``container.logs()``. For
        example, the ``stdout`` and ``stderr`` boolean arguments can be used to
        determine whether to stream stdout or stderr or both (the default).

    :raises TimeoutError:
        When the timeout value is reached before the logs have completed.
    """
    stream = container.logs(stream=True, **logs_kwargs)
    return stream_timeout(
        stream, timeout, 'Timeout waiting for container logs.')


def wait_for_logs_matching(container, matcher, timeout=10, encoding='utf-8',
                           **logs_kwargs):
    """
    Wait for matching log line(s) from the given container by streaming the
    container's stdout and/or stderr outputs.

    Each log line is decoded and any trailing whitespace is stripped before the
    line is matched.

    :param ~docker.models.containers.Container container:
        Container who's log lines to wait for.
    :param matcher:
        Callable that returns True once it has matched a decoded log line(s).
    :param timeout:
        Timeout value in seconds.
    :param encoding:
        Encoding to use when decoding container output to strings.
    :param logs_kwargs:
        Additional keyword arguments to pass to ``container.logs()``. For
        example, the ``stdout`` and ``stderr`` boolean arguments can be used to
        determine whether to stream stdout or stderr or both (the default).

    :returns:
        The final matching log line.
    :raises TimeoutError:
        When the timeout value is reached before matching log lines have been
        found.
    :raises RuntimeError:
        When all log lines have been consumed but matching log lines have not
        been found (the container must have stopped for its stream to have
        ended without error).
    """
    try:
        for line in stream_logs(container, timeout=timeout, **logs_kwargs):
            # Drop the trailing newline
            line = line.decode(encoding).rstrip()
            if matcher(line):
                return line
    except TimeoutError:
        raise TimeoutError('\n'.join([
            ('Timeout ({}s) waiting for logs matching {}.'.format(
                timeout, matcher)),
            'Last few log lines:',
            _last_few_log_lines(container),
        ]))

    raise RuntimeError('\n'.join([
        'Logs matching {} not found.'.format(matcher),
        'Last few log lines:',
        _last_few_log_lines(container),
    ]))



import os
import re

from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with open(os.path.join(HERE, *parts)) as f:
        return f.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string')


setup(
    name='seaworthy',
    version=find_version('seaworthy', '__init__.py'),
    license='BSD',
    url='https://github.com/praekeltfoundation/seaworthy',
    description='Test Docker container images',
    author='Praekelt.org SRE team',
    author_email='sre@praekelt.org',
    long_description=read('README.rst'),
    packages=find_packages(),
    install_requires=[
        'attrs>=17.4.0',
        'docker>=3.6,<4',
        'hyperlink',
        'requests',
    ],
    extras_require={
        'pytest': [
            'pytest>=3.0.0',
        ],
        'testtools': [
            'testtools',
        ],
        'test': [
            'pytest>=3.0.0',
            'responses',
            'testtools',
        ],
        'test-core': [
            'responses',
        ],
        'docstest': [
            'doc8',
            'readme_renderer',
            'Sphinx>=1.7,<1.8',
            'sphinx_rtd_theme',
        ],
        'pep8test': [
            'flake8',
            'flake8-import-order',
            'pep8-naming',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Pytest',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing',
    ],
    entry_points={
        'pytest11': ['seaworthy = seaworthy.pytest'],
    },
)



"""
These tests use the ``pytester`` plugin to run tests in a separate process.
Because of this, some Docker state can conflict between the main test process
and these new processes. For this reason, these tests have been separated from
the others. Note that these tests produce no coverage data.
https://docs.pytest.org/en/3.2.1/writing_plugins.html#testing-plugins
"""
from seaworthy.checks import docker_client
from seaworthy.helpers import fetch_images
from seaworthy.pytest.checks import dockertest


IMG = 'nginx:alpine'


def setup_module():
    with docker_client() as client:
        fetch_images(client, [IMG])


@dockertest()
class TestDockerHelperFixture:
    def test_fixture(self, testdir):
        """
        The ``docker_helper`` fixture should automatically be in scope for a
        project that uses this pytest plugin.  When the ``DockerHelper`` is
        passed to a test function using a fixture, it should be possible to
        create, start, stop, and remove a container using the helper.
        """
        testdir.makepyfile("""
            from docker.errors import NotFound
            import pytest

            def test_container_basics(docker_helper):
                container = docker_helper.containers.create('test', '{}')
                container.start()
                container.reload()

                assert container.status == 'running'

                docker_helper.containers.remove(container, force=True)

                with pytest.raises(NotFound):
                    container.reload()
        """.format(IMG))

        result = testdir.runpytest()
        result.assert_outcomes(passed=1)


@dockertest()
class TestImageFetchFixtureFunc:
    def test_fixture(self, testdir):
        """
        When the fixture is used in a test, the image passed to the test
        should have the right tag.
        """
        testdir.makeconftest("""
            from seaworthy.pytest.fixtures import image_fetch_fixture

            fixture = image_fetch_fixture('busybox', name='image')
        """.format(IMG))

        testdir.makepyfile("""
            def test_image_fetch(image):
                assert 'busybox:latest' in image.tags
        """)


@dockertest()
class TestResourceFixtureFunc:
    def test_container_fixture(self, testdir):
        """
        When the fixture is used in a test, the container passed to the test
        function should be running.
        """
        testdir.makeconftest("""
            from seaworthy.definitions import ContainerDefinition

            fixture = (ContainerDefinition(name='test', image='{}')
                       .pytest_fixture('container'))
        """.format(IMG))

        testdir.makepyfile("""
            def test_create_container(container):
                assert container.inner().status == 'running'
        """)

        result = testdir.runpytest()
        result.assert_outcomes(passed=1)

    def test_dependencies(self, testdir):
        """
        When the fixture is used in a test, and the fixture has a dependent
        fixture, the dependent fixture should be setup.
        """
        testdir.makeconftest("""
            from seaworthy.definitions import (
                ContainerDefinition, VolumeDefinition)

            class Container(ContainerDefinition):
                def __init__(self, volume, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.volume = volume

            volume = VolumeDefinition('foo')
            volume_fixture = volume.pytest_fixture('foo_volume')
            container_fixture = (
                Container(volume, name='test', image='{}').pytest_fixture(
                    'container', dependencies=('foo_volume',)))
        """.format(IMG))

        testdir.makepyfile("""
            def test_create_container(container):
                assert container.inner().status == 'running'
                assert container.volume.created
        """)

        result = testdir.runpytest()
        result.assert_outcomes(passed=1)


@dockertest()
class TestCleanContainerFixturesFunc:
    def test_fixture(self, testdir):
        """
        When the fixture is used in a test, it should be cleaned when the test
        function is marked to be cleaned. The container passed to the test
        function should be running.
        """
        testdir.makeconftest("""
            from seaworthy.definitions import ContainerDefinition
            from seaworthy.pytest.fixtures import clean_container_fixtures


            class CleanableContainer(ContainerDefinition):
                def __init__(self):
                    super().__init__(name='test', image='{}')
                    self.cleaned = False

                def clean(self):
                    self.cleaned = True

                def was_cleaned(self):
                    cleaned = self.cleaned
                    self.cleaned = False
                    return cleaned


            f1, f2 = CleanableContainer().pytest_clean_fixtures('cleanable')
        """.format(IMG))

        testdir.makepyfile("""
            import pytest


            def test_dirty(cleanable):
                assert not cleanable.was_cleaned()
                assert cleanable.inner().status == 'running'

            @pytest.mark.clean_cleanable
            def test_clean(cleanable):
                assert cleanable.was_cleaned()
                assert cleanable.inner().status == 'running'
        """)

        result = testdir.runpytest()
        result.assert_outcomes(passed=2)



"""
Checks and test decorators for skipping tests that require Docker to be
present.
"""

import unittest
from contextlib import contextmanager

import docker

from requests.exceptions import ConnectionError


@contextmanager
def docker_client():
    """
    A context manager that creates and cleans up a Docker API client.

    In most cases, it's better to use :class:`~seaworthy.helpers.DockerHelper`
    instead.
    """
    client = docker.client.from_env()
    yield client
    client.api.close()


def docker_available():
    """
    Check if Docker is available and responsive.
    """
    with docker_client() as client:
        try:
            return client.ping()
        except ConnectionError:  # pragma: no cover
            return False


def dockertest():
    """
    Skip tests that require Docker to be available.

    This is a function that returns a decorator so that we don't run arbitrary
    Docker client code on import. This implementation only works with tests
    based on :class:`unittest.TestCase`. If you're using pytest, you probably
    want :func:`seaworthy.pytest.dockertest` instead.
    """
    return unittest.skipUnless(docker_available(), 'Docker not available.')
